from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class LabelRef(BaseModelObject):
    """Reference to a label."""

    external_id: str


class Relationship(BaseModelObject):
    """Base class for relationship with common fields."""

    external_id: str
    source_external_id: str
    source_type: Literal["asset", "timeSeries", "file", "event", "sequence"]
    target_external_id: str
    target_type: Literal["asset", "timeSeries", "file", "event", "sequence"]
    start_time: int | None = None
    end_time: int | None = None
    confidence: float | None = None
    data_set_id: int | None = None
    labels: list[LabelRef] | None = None


class RelationshipRequest(Relationship, UpdatableRequestResource):
    """Request resource for creating/updating relationships."""

    container_fields: ClassVar[frozenset[str]] = frozenset({"labels"})

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RelationshipResponse(Relationship, ResponseResource[RelationshipRequest]):
    """Response resource for relationships."""

    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[RelationshipRequest]:
        return RelationshipRequest



from typing import ClassVar, Literal

from pydantic import field_serializer

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
)

from .data_modeling import ViewReference
from .instance_api import TypedNodeIdentifier, TypedViewReference, WrappedInstanceRequest, WrappedInstanceResponse

RESOURCE_VIEW_MAPPING_SPACE: Literal["cognite_migration"] = "cognite_migration"
RESOURCE_MAPPING_VIEW_ID = TypedViewReference(
    space=RESOURCE_VIEW_MAPPING_SPACE, external_id="ResourceViewMapping", version="v1"
)


class ResourceViewMapping(BaseModelObject):
    resource_type: str
    view_id: ViewReference
    property_mapping: dict[str, str]

    @field_serializer("view_id")
    def serialize_view_id(self, view_id: ViewReference) -> dict:
        return {**view_id.dump(), "type": "view"}


class ResourceViewMappingRequest(WrappedInstanceRequest, ResourceViewMapping):
    VIEW_ID: ClassVar[TypedViewReference] = RESOURCE_MAPPING_VIEW_ID
    space: Literal["cognite_migration"] = RESOURCE_VIEW_MAPPING_SPACE
    instance_type: Literal["node"] = "node"

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)


class ResourceViewMappingResponse(WrappedInstanceResponse[ResourceViewMappingRequest], ResourceViewMapping):
    VIEW_ID: ClassVar[TypedViewReference] = RESOURCE_MAPPING_VIEW_ID
    space: Literal["cognite_migration"] = RESOURCE_VIEW_MAPPING_SPACE
    instance_type: Literal["node"] = "node"

    @classmethod
    def request_cls(cls) -> type[ResourceViewMappingRequest]:
        return ResourceViewMappingRequest

    def as_request_resource(self) -> ResourceViewMappingRequest:
        return ResourceViewMappingRequest.model_validate(self.dump(context="toolkit"), extra="ignore")



import platform
import sys
from pathlib import Path

from rich import print
from rich.table import Table

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml, _read_toml
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import clean_name
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.plugins import Plugins
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._version import __version__


class AboutCommand(ToolkitCommand):
    def execute(self, cwd: Path) -> None:
        # Version information
        print(f"\n[bold cyan]Cognite Toolkit[/bold cyan] version: [yellow]{__version__}[/yellow]")
        print(f"Python version: {sys.version.split()[0]}")
        print(f"Platform: {platform.system()} {platform.release()}")

        # Check for cdf.toml in the current directory
        cdf_toml_path = cwd / CDFToml.file_name

        if cdf_toml_path.exists():
            print(f"\n[bold green]Configuration file found:[/bold green] {cdf_toml_path}")

            cdf_toml = CDFToml.load(cwd)

            # We need to read the raw TOML to get original key names for plugins and alpha flags
            raw_toml = _read_toml(cdf_toml_path)

            self._check_unrecognized_sections(raw_toml)
            self._display_plugins(cdf_toml, raw_toml)
            self._display_alpha_flags(cdf_toml, raw_toml)
            self._display_additional_config(cdf_toml)

        else:
            # Search for cdf.toml in subdirectories
            found_files = self._search_cdf_toml(cwd)

            if found_files:
                print(f"\n[bold yellow]No cdf.toml found in current directory:[/bold yellow] {cwd}")
                print("\n[bold]Found cdf.toml files in subdirectories:[/bold]")
                for file in found_files:
                    rel_path = file.relative_to(cwd)
                    print(f"  • {rel_path}")
                print(f"\n[bold cyan]Hint:[/bold cyan] Move one of these files to {cwd} or navigate to its directory.")
            else:
                print("\n[bold yellow]No cdf.toml found[/bold yellow] in current directory or subdirectories.")
                print(f"Current directory: {cwd}")
                print("\n[bold cyan]Hint:[/bold cyan] Run [yellow]cdf init[/yellow] to create a new project.")

    def _check_unrecognized_sections(self, raw_toml: dict) -> None:
        """Check for unrecognized tables in cdf.toml and warn about them."""
        # Valid top-level tables in cdf.toml
        valid_tables = {"cdf", "modules", "alpha_flags", "feature_flags", "plugins", "library"}

        # Filter out empty keys, whitespace-only keys, and check for unrecognized tables
        unrecognized_tables = [key for key in raw_toml.keys() if key and key.strip() and key not in valid_tables]

        if unrecognized_tables:
            print()

        for table in unrecognized_tables:
            # Try to find a matching valid table by stripping non-alphabetical characters
            suggestion = self._find_similar_table(table, valid_tables)

            message = f"Table '{table}' in cdf.toml is not recognized and will have no effect."
            if suggestion:
                message += f" Did you mean '{suggestion}'?"

            self.warn(MediumSeverityWarning(message))

    @staticmethod
    def _find_similar_table(unrecognized: str, valid_tables: set[str]) -> str | None:
        """Find a similar valid table by comparing alphabetical characters only.

        Returns None if the unrecognized table is already valid or if no similar match is found.
        """
        # If it's already a valid table, return None (no suggestion needed)
        if unrecognized in valid_tables:
            return None

        # Keep only alphabetical characters and lowercase
        normalized_unrecognized = "".join(c for c in unrecognized if c.isalpha()).lower()

        # First, try exact match (after normalization)
        for valid in valid_tables:
            normalized_valid = "".join(c for c in valid if c.isalpha()).lower()
            if normalized_unrecognized == normalized_valid:
                return valid

        # If no match, check for singular/plural variations (missing 's')
        for valid in valid_tables:
            normalized_valid = "".join(c for c in valid if c.isalpha()).lower()

            # Check if adding 's' to unrecognized matches valid (e.g., "plugin" -> "plugins")
            if normalized_unrecognized + "s" == normalized_valid:
                return valid

        return None

    def _display_plugins(self, cdf_toml: CDFToml, raw_toml: dict) -> None:
        """Display all available plugins and their status."""
        table = Table(title="Plugins", show_header=True)
        table.add_column("Plugin", justify="left", style="cyan")
        table.add_column("Status", justify="center")
        table.add_column("Description", justify="left")

        # Track which plugins we've seen
        seen_plugins = set()

        # Show all plugins from the enum
        for plugin in Plugins:
            plugin_name = plugin.value.name
            cleaned_key = clean_name(plugin_name)
            seen_plugins.add(cleaned_key)

            is_enabled = cdf_toml.plugins.get(cleaned_key, False)
            if is_enabled:
                status = "[green]✓ enabled[/green]"
            else:
                status = "[dim]○ disabled[/dim]"

            table.add_row(plugin_name, status, plugin.value.description)

        print()
        print(table)

        # Show any unrecognized plugins from cdf.toml using original key names
        raw_plugins = raw_toml.get("plugins", {})
        unrecognized = []
        for original_key, value in raw_plugins.items():
            cleaned_key = clean_name(original_key)
            if cleaned_key not in seen_plugins:
                unrecognized.append((original_key, value))

        for original_key, is_enabled in unrecognized:
            status = "enabled" if is_enabled else "disabled"
            self.warn(
                LowSeverityWarning(f"Plugin '{original_key}' in cdf.toml is not recognized and will have no effect.")
            )

    def _display_alpha_flags(self, cdf_toml: CDFToml, raw_toml: dict) -> None:
        """Display available alpha flags and their status."""
        table = Table(title="Alpha Flags", show_header=True)
        table.add_column("Flag", justify="left", style="yellow")
        table.add_column("Status", justify="center")
        table.add_column("Description", justify="left")

        # Track which flags we've seen
        seen_flags = set()

        # Show flags from the enum that are either enabled or visible
        for flag in Flags:
            cleaned_key = clean_name(flag.name)
            seen_flags.add(cleaned_key)

            is_enabled = cdf_toml.alpha_flags.get(cleaned_key, False)

            # Only show if enabled or visible
            if is_enabled or flag.value.visible:
                # Convert enum name to kebab-case for display
                display_name = flag.name.lower().replace("_", "-")

                if is_enabled:
                    status = "[green]✓ enabled[/green]"
                else:
                    status = "[dim]○ disabled[/dim]"

                table.add_row(display_name, status, flag.value.description)

        print()
        print(table)

        # Show any unrecognized flags from cdf.toml using original key names
        raw_flags = raw_toml.get("alpha_flags", {})
        unrecognized = []
        for original_key, value in raw_flags.items():
            cleaned_key = clean_name(original_key)
            if cleaned_key not in seen_flags:
                unrecognized.append((original_key, value))

        for original_key, is_enabled in unrecognized:
            status = "enabled" if is_enabled else "disabled"
            self.warn(
                LowSeverityWarning(
                    f"Alpha flag '{original_key}' in cdf.toml is not recognized and will have no effect."
                )
            )

    def _display_additional_config(self, cdf_toml: CDFToml) -> None:
        """Display additional configuration information."""
        print("\n[bold]Additional Configuration:[/bold]")

        print(f"  Default environment: [cyan]{cdf_toml.cdf.default_env}[/cyan]")

        if cdf_toml.cdf.has_user_set_default_org:
            print(f"  Default organization dir: [cyan]{cdf_toml.cdf.default_organization_dir}[/cyan]")

        if cdf_toml.cdf.file_encoding:
            print(f"  File encoding: [cyan]{cdf_toml.cdf.file_encoding}[/cyan]")

        print(f"  Modules version: [cyan]{cdf_toml.modules.version}[/cyan]")

        if cdf_toml.libraries:
            print(f"  Configured libraries: [cyan]{len(cdf_toml.libraries)}[/cyan]")
            for lib_name, lib_config in cdf_toml.libraries.items():
                print(f"    • {lib_name}: [dim]{lib_config.url}[/dim]")

    def _search_cdf_toml(self, cwd: Path) -> list[Path]:
        """Search for cdf.toml files in immediate subdirectories (one level down)."""
        try:
            return sorted(
                [potential_file for potential_file in cwd.glob(f"*/{CDFToml.file_name}") if potential_file.is_file()]
            )
        except PermissionError:
            return []



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .base import ToolkitResource


class DataProductYAML(ToolkitResource):
    external_id: str = Field(
        description="User-defined unique identifier for the Data Product.",
        min_length=1,
        max_length=100,
        pattern=r"^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$",
    )
    name: str = Field(
        description="Human-readable name of the Data Product.",
        min_length=1,
        max_length=50,
    )
    schema_space: str | None = Field(
        default=None,
        description="The schema space where the data product's data model and views are located. Defaults to the data product's externalId.",
    )
    description: str | None = Field(
        default=None,
        description="A description of the Data Product.",
        max_length=200,
    )
    is_governed: bool = Field(
        default=False,
        description="Indicates whether the data product follows governance policies and standards.",
    )
    tags: list[str] | None = Field(
        default=None, description="A list of distinct tags for categorization and filtering.", max_length=10
    )

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



from typing import ClassVar

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId


class DataSet(BaseModelObject):
    external_id: str | None = None
    name: str | None = None
    description: str | None = None
    metadata: Metadata | None = None
    write_protected: bool | None = None

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert DataSet to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class DataSetRequest(DataSet, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"write_protected"})


class DataSetResponse(DataSet, ResponseResource[DataSetRequest]):
    id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[DataSetRequest]:
        return DataSetRequest



"""Views API for managing CDF data modeling views.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Views/operation/ApplyViews
"""

from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ViewFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ViewReference,
    ViewReferenceNoVersion,
    ViewRequest,
    ViewResponse,
)


class ViewsAPI(CDFResourceAPI[ViewResponse]):
    """API for managing CDF data modeling views.

    Views use an apply/upsert pattern for create and update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/models/views", item_limit=100),
                "retrieve": Endpoint(method="POST", path="/models/views/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/models/views/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/models/views", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[ViewResponse]:
        return PagedResponse[ViewResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ViewRequest]) -> list[ViewResponse]:
        """Apply (create or update) views in CDF.

        Args:
            items: List of ViewRequest objects to apply.

        Returns:
            List of applied ViewResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def update(self, items: Sequence[ViewRequest]) -> list[ViewResponse]:
        """Apply (create or update) views in CDF.

        Args:
            items: List of ViewRequest objects to apply.
        Returns:
            List of applied ViewResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def retrieve(
        self, items: Sequence[ViewReferenceNoVersion], include_inherited_properties: bool = True
    ) -> list[ViewResponse]:
        """Retrieve views from CDF.

        Args:
            items: List of ViewReference objects to retrieve.
            include_inherited_properties: Whether to include inherited properties.

        Returns:
            List of retrieved ViewResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", params={"includeInheritedProperties": include_inherited_properties}
        )

    def delete(self, items: Sequence[ViewReference]) -> None:
        """Delete views from CDF.

        Args:
            items: List of ViewReference objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: ViewFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ViewResponse]:
        """Get a page of views from CDF.

        Args:
            filter: ViewFilter to filter views.
            limit: Maximum number of views to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of ViewResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def iterate(
        self,
        filter: ViewFilter | None = None,
        limit: int | None = None,
    ) -> Iterable[list[ViewResponse]]:
        """Iterate over all views in CDF.

        Args:
            filter: ViewFilter to filter views.
            limit: Maximum total number of views to return.

        Returns:
            Iterable of lists of ViewResponse objects.
        """
        return self._iterate(
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def list(
        self,
        filter: ViewFilter | None = None,
        limit: int | None = None,
    ) -> list[ViewResponse]:
        """List all views in CDF.

        Args:
            filter: ViewFilter to filter views.
            limit: Maximum total number of views to return.

        Returns:
            List of ViewResponse objects.
        """
        return self._list(limit=limit, params=filter.dump() if filter else None)



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.transformation_schedule import (
    TransformationScheduleRequest,
    TransformationScheduleResponse,
)


class TransformationSchedulesAPI(CDFResourceAPI[TransformationScheduleResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/transformations/schedules", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/transformations/schedules/byids", item_limit=1000),
                "update": Endpoint(method="POST", path="/transformations/schedules/update", item_limit=1000),
                "delete": Endpoint(method="POST", path="/transformations/schedules/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/transformations/schedules", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[TransformationScheduleResponse]:
        return PagedResponse[TransformationScheduleResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[TransformationScheduleRequest]) -> list[TransformationScheduleResponse]:
        """Schedule transformations with the specified configurations.

        Args:
            items: List of TransformationScheduleRequest objects to create.
        Returns:
            List of created TransformationScheduleResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False
    ) -> list[TransformationScheduleResponse]:
        """Retrieve transformation schedules from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved TransformationScheduleResponse objects.
        """
        return self._request_item_response(
            items,
            method="retrieve",
            extra_body={"ignoreUnknownIds": ignore_unknown_ids},
        )

    def update(
        self, items: Sequence[TransformationScheduleRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[TransformationScheduleResponse]:
        """Update transformation schedules in CDF.

        Args:
            items: List of TransformationScheduleRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated TransformationScheduleResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Unschedule transformations from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[TransformationScheduleResponse]:
        """Fetch a page of transformation schedules from CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of TransformationScheduleResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
        )

    def iterate(
        self,
        limit: int | None = 100,
    ) -> Iterable[list[TransformationScheduleResponse]]:
        """Iterate over all transformation schedules in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of TransformationScheduleResponse objects.
        """
        return self._iterate(limit=limit)

    def list(
        self,
        limit: int | None = 100,
    ) -> list[TransformationScheduleResponse]:
        """List all transformation schedules in CDF.

        Returns:
            List of TransformationScheduleResponse objects.
        """
        return self._list(limit=limit)



from typing import Annotated

from pydantic import Field, TypeAdapter

from ._auth import (
    BasicAuthenticationRequest,
    BasicAuthenticationResponse,
    ClientCredentialAuthenticationRequest,
    ClientCredentialAuthenticationResponse,
    HTTPBasicAuthenticationRequest,
    HTTPBasicAuthenticationResponse,
    ScramShaAuthenticationRequest,
    ScramShaAuthenticationResponse,
)
from ._certificate import AuthCertificateRequest, CACertificateRequest, CertificateResponse
from ._eventhub import EventHubSourceRequest, EventHubSourceResponse
from ._kafka import KafkaBroker, KafkaSourceRequest, KafkaSourceResponse
from ._mqtt import MQTTSourceRequest, MQTTSourceResponse
from ._rest import RESTSourceRequest, RESTSourceResponse

HostedExtractorSourceRequestUnion = Annotated[
    KafkaSourceRequest | EventHubSourceRequest | MQTTSourceRequest | RESTSourceRequest,
    Field(discriminator="type"),
]

HostedExtractorSourceRequest: TypeAdapter[HostedExtractorSourceRequestUnion] = TypeAdapter(
    HostedExtractorSourceRequestUnion
)

HostedExtractorSourceResponseUnion = Annotated[
    KafkaSourceResponse | EventHubSourceResponse | MQTTSourceResponse | RESTSourceResponse,
    Field(discriminator="type"),
]

HostedExtractorSourceResponse: TypeAdapter[HostedExtractorSourceResponseUnion] = TypeAdapter(
    HostedExtractorSourceResponseUnion
)


__all__ = [
    "AuthCertificateRequest",
    "BasicAuthenticationRequest",
    "BasicAuthenticationResponse",
    "CACertificateRequest",
    "CertificateResponse",
    "ClientCredentialAuthenticationRequest",
    "ClientCredentialAuthenticationResponse",
    "EventHubSourceRequest",
    "EventHubSourceResponse",
    "HTTPBasicAuthenticationRequest",
    "HTTPBasicAuthenticationResponse",
    "HostedExtractorSourceRequest",
    "HostedExtractorSourceResponse",
    "KafkaBroker",
    "KafkaSourceRequest",
    "KafkaSourceResponse",
    "MQTTSourceRequest",
    "MQTTSourceResponse",
    "RESTSourceRequest",
    "RESTSourceResponse",
    "ScramShaAuthenticationRequest",
    "ScramShaAuthenticationResponse",
]



from collections.abc import Hashable, Iterable, Sequence
from graphlib import CycleError, TopologicalSorter
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, LocationFiltersAcl

from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    DataModelReference,
    SpaceReference,
    ViewReference,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.location_filter import (
    LocationFilterRequest,
    LocationFilterResponse,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ResourceRetrievalError, ToolkitCycleError
from cognite_toolkit._cdf_tk.resource_classes import LocationYAML
from cognite_toolkit._cdf_tk.utils import in_dict, quote_int_value_by_key_in_yaml, safe_read
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier

from .classic import AssetCRUD, SequenceCRUD
from .data_organization import DataSetsCRUD
from .datamodel import DataModelCRUD, SpaceCRUD, ViewCRUD
from .file import FileMetadataCRUD
from .timeseries import TimeSeriesCRUD


@final
class LocationFilterCRUD(ResourceCRUD[ExternalId, LocationFilterRequest, LocationFilterResponse]):
    folder_name = "locations"
    resource_cls = LocationFilterResponse
    resource_write_cls = LocationFilterRequest
    yaml_cls = LocationYAML
    dependencies = frozenset(
        {
            AssetCRUD,
            DataSetsCRUD,
            DataModelCRUD,
            SpaceCRUD,
            ViewCRUD,
            SequenceCRUD,
            FileMetadataCRUD,
            TimeSeriesCRUD,
        }
    )
    kind = "LocationFilter"
    _doc_base_url = "https://api-docs.cogheim.net/redoc/#tag/"
    _doc_url = "Location-Filters/operation/createLocationFilter"

    subfilter_names = ("assets", "events", "files", "timeseries", "sequences")

    @property
    def display_name(self) -> str:
        return "location filters"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[LocationFilterRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        # Todo: Specify space ID scopes:

        actions = (
            [LocationFiltersAcl.Action.Read]
            if read_only
            else [LocationFiltersAcl.Action.Read, LocationFiltersAcl.Action.Write]
        )

        return LocationFiltersAcl(
            actions=actions,
            scope=LocationFiltersAcl.Scope.All(),
            allow_unknown=True,
        )

    @classmethod
    def get_id(cls, item: LocationFilterRequest | LocationFilterResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("LocationFilter must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return {"externalId": id.external_id}

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return id.external_id

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> LocationFilterRequest:
        if parent_external_id := resource.pop("parentExternalId", None):
            # This is a workaround: when the parentExternalId cannot be resolved because the parent
            # hasn't been created yet, we save it so that we can try again "later"
            try:
                resource["parentId"] = self.client.lookup.location_filters.id(parent_external_id, is_dry_run)
            except ResourceRetrievalError:
                resource["parentId"] = -1
            # Store the parent external ID for topological sorting and late look up.
            resource["parentExternalId"] = parent_external_id

        if "assetCentric" not in resource:
            return LocationFilterRequest.model_validate(resource)
        asset_centric = resource["assetCentric"]
        if data_set_external_ids := asset_centric.pop("dataSetExternalIds", None):
            asset_centric["dataSetIds"] = self.client.lookup.data_sets.id(
                data_set_external_ids, is_dry_run, allow_empty=True
            )
        for subfilter_name in self.subfilter_names:
            subfilter = asset_centric.get(subfilter_name, {})
            if data_set_external_ids := subfilter.pop("dataSetExternalIds", []):
                asset_centric[subfilter_name]["dataSetIds"] = self.client.lookup.data_sets.id(
                    data_set_external_ids,
                    is_dry_run,
                    allow_empty=True,
                )

        return LocationFilterRequest.model_validate(resource)

    def dump_resource(self, resource: LocationFilterResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        if parent_id := dumped.pop("parentId", None):
            dumped["parentExternalId"] = self.client.lookup.location_filters.external_id(parent_id)
        if dumped.get("dataModelingType") == "HYBRID" and local is not None and "dataModelingType" not in local:
            # Default set on server side
            dumped.pop("dataModelingType")
        if "assetCentric" not in dumped:
            return dumped
        asset_centric = dumped["assetCentric"]
        if data_set_ids := asset_centric.pop("dataSetIds", None):
            asset_centric["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(data_set_ids)
        for subfilter_name in self.subfilter_names:
            subfilter = asset_centric.get(subfilter_name, {})
            if data_set_ids := subfilter.pop("dataSetIds", []):
                asset_centric[subfilter_name]["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(
                    data_set_ids
                )
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "assetCentric" or json_path == ("instanceSpaces",):
            return diff_list_hashable(local, cdf)
        elif json_path in [("dataModels",), ("views",), ("scene",)]:
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    @classmethod
    def topological_sort(cls, items: Sequence[LocationFilterRequest]) -> list[LocationFilterRequest]:
        """Sorts the location filters in topological order based on their parent-child relationships."""
        location_by_id: dict[str, LocationFilterRequest] = {item.external_id: item for item in items}
        dependencies: dict[str, set[str]] = {}
        for item_id, item in location_by_id.items():
            dependencies[item_id] = set()
            # If this item has a parent, add it as a dependency
            if item.parent_external_id and item.parent_external_id in location_by_id:
                dependencies[item_id].add(item.parent_external_id)
        try:
            return [
                location_by_id[item_id]
                for item_id in TopologicalSorter(dependencies).static_order()
                if item_id in location_by_id
            ]
        except CycleError as e:
            raise ToolkitCycleError(
                f"Cannot deploy location filters. Cycle detected {e.args} in the parent-child dependencies of the location filters.",
                *e.args[1:],
            ) from None

    def create(self, items: Sequence[LocationFilterRequest]) -> list[LocationFilterResponse]:
        created: list[LocationFilterResponse] = []
        # Note: the Location API does not support batch creation, so we need to do this one by one.
        # Furthermore, we could not do the parentExternalId->parentId lookup before the parent was created,
        # hence it may be deferred here.
        # Use topological sort to ensure parents are created before children
        for item in self.topological_sort(items):
            # These are set if lookup has been deferred
            if item.parent_external_id and item.parent_id == -1:
                item.parent_id = self.client.lookup.location_filters.id(item.parent_external_id)
            created.extend(self.client.tool.location_filters.create([item]))
        return created

    def retrieve(self, external_ids: Sequence[ExternalId]) -> list[LocationFilterResponse]:
        # Use flat=True to get all locations in a flat list
        all_locations = self.client.tool.location_filters.list(flat=True)
        external_id_set = {ext_id.external_id for ext_id in external_ids}
        return [loc for loc in all_locations if loc.external_id in external_id_set]

    def update(self, items: Sequence[LocationFilterRequest]) -> list[LocationFilterResponse]:
        all_locations = self.client.tool.location_filters.list(flat=True)
        ids = {loc.external_id: loc.id for loc in all_locations}
        # Set the id on each item before updating
        for item in items:
            item.id = ids[item.external_id]
        return self.client.tool.location_filters.update(items)

    def delete(self, external_ids: Sequence[ExternalId]) -> int:
        locations = self.retrieve(external_ids)
        if not locations:
            return 0
        ids = [InternalId(id=loc.id) for loc in locations]
        self.client.tool.location_filters.delete(ids)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[LocationFilterResponse]:
        for chunk in self.client.tool.location_filters.iterate(flat=True):
            yield from chunk

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "assetCentric" in item:
            asset_centric = item["assetCentric"]
            for data_set_external_id in asset_centric.get("dataSetExternalIds", []):
                yield DataSetsCRUD, ExternalId(external_id=data_set_external_id)
            for asset in asset_centric.get("assetSubtreeIds", []):
                if "externalId" in asset:
                    yield AssetCRUD, ExternalId(external_id=asset["externalId"])
            for subfilter_name in cls.subfilter_names:
                subfilter = asset_centric.get(subfilter_name, {})
                for data_set_external_id in subfilter.get("dataSetExternalIds", []):
                    yield DataSetsCRUD, ExternalId(external_id=data_set_external_id)
                for asset in subfilter.get("assetSubtreeIds", []):
                    if "externalId" in asset:
                        yield AssetCRUD, ExternalId(external_id=asset["externalId"])
        for view in item.get("views", []):
            if in_dict(["space", "externalId", "version"], view):
                yield (
                    ViewCRUD,
                    ViewReference(space=view["space"], external_id=view["externalId"], version=view["version"]),
                )
        for space in item.get("instanceSpaces", []):
            yield SpaceCRUD, SpaceReference(space=space)
        for data_model in item.get("dataModels", []):
            if in_dict(["space", "externalId", "version"], data_model):
                yield (
                    DataModelCRUD,
                    DataModelReference(
                        space=data_model["space"], external_id=data_model["externalId"], version=data_model["version"]
                    ),
                )



from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import Identifier


class SpaceReference(Identifier):
    space: str

    def __str__(self) -> str:
        return self.space


class ContainerReference(Identifier):
    type: Literal["container"] = Field("container", exclude=True)
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"

    def as_tuple(self) -> tuple[str, str]:
        return self.space, self.external_id


class ViewReferenceNoVersion(Identifier):
    type: Literal["view"] = Field("view", exclude=True)
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"


class ViewReference(ViewReferenceNoVersion):
    version: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(version={self.version})"


class DataModelReferenceNoVersion(Identifier):
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"


class DataModelReference(DataModelReferenceNoVersion):
    version: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(version={self.version})"


class NodeReference(Identifier):
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"


class EdgeReference(Identifier):
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}"


class ContainerDirectReference(Identifier):
    source: ContainerReference
    identifier: str

    def __str__(self) -> str:
        return f"{self.source!s}.{self.identifier}"


class ViewDirectReference(Identifier):
    source: ViewReference
    identifier: str

    def __str__(self) -> str:
        return f"{self.source!s}.{self.identifier}"


class ContainerIndexReference(ContainerReference):
    identifier: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(index={self.identifier})"


class ContainerConstraintReference(ContainerReference):
    identifier: str

    def __str__(self) -> str:
        return f"{self.space}:{self.external_id}(constraint={self.identifier})"


# Todo: Temporary put here to avoid circular imports.


class DatapointSubscriptionTimeSeriesId(Identifier):
    external_id: str | None = None
    id: int | None = None
    instance_id: NodeReference | None = None

    def __str__(self) -> str:
        if self.external_id is not None:
            return f"externalId='{self.external_id}'"
        elif self.id is not None:
            return f"id={self.id}"
        elif self.instance_id is not None:
            return f"instanceId={self.instance_id!s}"
        else:
            return "undefined"


class InstanceId(Identifier):
    instance_id: NodeReference

    def __str__(self) -> str:
        return f"instanceId='{self.instance_id}'"



from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import RespaceCommand


class RespaceApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("plan")(self.respace_plan)
        self.command("execute")(self.respace_execute)

    @staticmethod
    def main(ctx: typer.Context) -> None:
        """Commands for respacing (moving) nodes between CDF spaces."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf data respace --help[/] for more information.")

    @staticmethod
    def respace_plan(
        csv_file: Annotated[
            Path,
            typer.Argument(
                help="Path to CSV file with nodes to respace. Expected columns: sourceSpace, externalId, targetSpace.",
                exists=True,
                file_okay=True,
                dir_okay=False,
                resolve_path=True,
            ),
        ],
        output_file: Annotated[
            Path,
            typer.Option(
                "--output-file",
                "-o",
                help="Path for the output plan file.",
            ),
        ] = Path("respace_plan.json"),
    ) -> None:
        """Generate a respace plan by analyzing nodes and their dependencies."""
        cmd = RespaceCommand()
        cmd.run(lambda: cmd.plan(csv_file=csv_file, output_file=output_file))

    @staticmethod
    def respace_execute(
        plan_file: Annotated[
            Path,
            typer.Argument(
                help="Path to the respace plan JSON generated by 'plan' command.",
                exists=True,
                file_okay=True,
                dir_okay=False,
                resolve_path=True,
            ),
        ],
        backup_dir: Annotated[
            Path,
            typer.Option(
                "--backup-dir",
                "-b",
                help="Directory to store backup data before migration.",
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="Preview changes without executing.",
            ),
        ] = False,
    ) -> None:
        """Execute a respace plan, migrating nodes between spaces."""
        cmd = RespaceCommand()
        cmd.run(lambda: cmd.execute(plan_file=plan_file, backup_dir=backup_dir, dry_run=dry_run))



from typing import Any

from pydantic import JsonValue, field_serializer, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, Identifier, RequestResource, ResponseResource

from .data_modeling._references import ContainerReference


class RecordIdentifier(Identifier):
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"Record({self.space}, {self.external_id})"


class RecordSource(BaseModelObject):
    source: ContainerReference
    properties: dict[str, JsonValue]

    @field_serializer("source", mode="plain")
    def serialize_source(self, value: ContainerReference) -> Any:
        return {**value.dump(), "type": value.type}


class RecordRequest(RequestResource):
    """A record request resource for ingesting into a stream."""

    space: str
    external_id: str
    sources: list[RecordSource]

    def as_id(self) -> RecordIdentifier:
        return RecordIdentifier(space=self.space, external_id=self.external_id)


class RecordResponse(ResponseResource[RecordRequest]):
    """A record response from the API."""

    space: str
    external_id: str
    properties: dict[ContainerReference, dict[str, JsonValue]] | None = None

    @field_validator("properties", mode="before")
    def parse_reference(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        parsed: dict[ContainerReference, dict[str, JsonValue]] = {}
        for space, inner_dict in value.items():
            if isinstance(space, ContainerReference):
                parsed[space] = inner_dict
                continue
            if not isinstance(inner_dict, dict) or not isinstance(space, str):
                raise ValueError(
                    f"Invalid properties format expected dict[str, dict[...]], "
                    f"got: dict[{type(space).__name__}, {type(inner_dict).__name__}]"
                )
            for container_external_id, props in inner_dict.items():
                source_ref = ContainerReference(space=space, external_id=container_external_id)  # pyright: ignore[reportCallIssue]
                parsed[source_ref] = props
        return parsed

    @field_serializer("properties", mode="plain")
    def serialize_properties(
        self, value: dict[ContainerReference, dict[str, JsonValue]] | None
    ) -> dict[str, dict[str, dict[str, JsonValue]]] | None:
        if value is None:
            return None
        serialized: dict[str, dict[str, dict[str, JsonValue]]] = {}
        for source_ref, props in value.items():
            space = source_ref.space
            if space not in serialized:
                serialized[space] = {}
            serialized[space][source_ref.external_id] = props
        return serialized

    @classmethod
    def request_cls(cls) -> type[RecordRequest]:
        return RecordRequest

    def as_request_resource(self) -> RecordRequest:
        dumped = self.dump()
        if self.properties:
            dumped["sources"] = [
                RecordSource(source=source_ref, properties=props) for source_ref, props in self.properties.items()
            ]
        return RecordRequest.model_validate(dumped, extra="ignore")



"""Principals API for managing CDF principals.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Principals
"""

from collections.abc import Iterable, Sequence

from pydantic import TypeAdapter

from cognite_toolkit._cdf_tk.client.api.project import ProjectAPI
from cognite_toolkit._cdf_tk.client.cdf_client.api import CDFResourceAPI, Endpoint
from cognite_toolkit._cdf_tk.client.cdf_client.responses import PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import (
    ExternalId,
    PrincipalId,
    PrincipalLoginId,
)
from cognite_toolkit._cdf_tk.client.resource_classes.principal import LoginSession, Principal, PrincipalType
from cognite_toolkit._cdf_tk.utils.useful_types import PrimitiveType


class PrincipalLoginSessionsAPI(CDFResourceAPI[LoginSession]):
    """API for the Cognite Principal Login Sessions endpoints.

    This API is used to manage login sessions for principals. It is not project-scoped.
    """

    def __init__(self, http_client: HTTPClient, project_api: ProjectAPI) -> None:
        self._revoke = Endpoint(method="POST", path="/orgs/{org}/principals/{principal}/sessions/revoke", item_limit=10)
        super().__init__(
            http_client,
            method_endpoint_map={
                "list": Endpoint(method="GET", path="/orgs/{org}/principals/{principal}/sessions", item_limit=100),
                # Misusing retrieve here to use the generic _request_item_response method.
                "retrieve": self._revoke,
            },
        )
        self._project_api = project_api

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint."""
        return self._http_client.config.create_auth_url(path)

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[LoginSession]:
        return PagedResponse[LoginSession].model_validate_json(response.body)

    def revoke(self, items: list[PrincipalLoginId]) -> None:
        """Revoke login sessions for a principal.

        Args:
            items: A list of PrincipalLoginId objects representing the sessions to revoke.
        """
        for principal, principal_items in self._group_items_by_text_field(items, "principal").items():
            path = self._revoke.path.format(org=self._project_api.get_organization_id(), principal=principal)
            self._request_no_response(items, "retrieve", endpoint=path)

    def paginate(self, principal_id: str, limit: int = 10, cursor: str | None = None) -> PagedResponse[LoginSession]:
        """Paginate through all login sessions for a principal.

        Args:
            principal_id: The ID of the principal to list sessions for.
            limit: The maximum number of items to return per page. Default is 10.
        Returns:
            A sequence of LoginSession objects.
        """
        path = self._method_endpoint_map["list"].path.format(
            org=self._project_api.get_organization_id(), principal=principal_id
        )
        result = self._paginate(limit, cursor, endpoint_path=path)
        for item in result.items:
            item.principal = principal_id
        return result

    def iterate(
        self, principal_id: str, limit: int | None = 10, cursor: str | None = None
    ) -> Iterable[list[LoginSession]]:
        """Iterate through all login sessions for a principal."""
        endpoint_path = self._method_endpoint_map["list"].path.format(
            org=self._project_api.get_organization_id(), principal=principal_id
        )
        for items in self._iterate(endpoint_path=endpoint_path, limit=limit, cursor=cursor):
            for item in items:
                item.principal = principal_id
            yield items

    def list(self, principal_id: str, limit: int | None = 10) -> list[LoginSession]:
        """List login sessions for a principal.

        Args:
            principal_id: The ID of the principal to list sessions for.
            limit: The maximum number of items to return. Default is 10.
        Returns:
            A list of LoginSession objects.
        """
        endpoint_path = self._method_endpoint_map["list"].path.format(
            org=self._project_api.get_organization_id(), principal=principal_id
        )
        results = self._list(endpoint_path=endpoint_path, limit=limit)
        for item in results:
            item.principal = principal_id
        return results


class PrincipalsAPI(CDFResourceAPI[Principal]):
    """API for the Cognite Principals endpoints.

    Unlike most CDF APIs, the Principals API is not project-scoped.
    The `me` endpoint uses `/api/v1/principals/me`, and all other
    endpoints use `/api/v1/orgs/{org}/principals/...`.
    """

    def __init__(self, http_client: HTTPClient, project_api: ProjectAPI) -> None:
        super().__init__(
            http_client,
            method_endpoint_map={
                "list": Endpoint(method="GET", path="/orgs/{org}/principals", item_limit=100),
                "retrieve": Endpoint(method="POST", path="/orgs/{org}/principals/byids", item_limit=100),
            },
        )
        self._me_endpoint = Endpoint(method="GET", path="/principals/me", item_limit=1)
        self._project_api = project_api
        self.login_sessions = PrincipalLoginSessionsAPI(http_client, project_api)

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint."""
        return self._http_client.config.create_auth_url(path)

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[Principal]:
        return PagedResponse[Principal].model_validate_json(response.body)

    def me(self) -> Principal:
        """Get the current caller's principal information.

        Returns:
            The principal (ServiceAccountPrincipal or UserPrincipal) that issued the request.
        """
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._http_client.config.create_auth_url(self._me_endpoint.path),
                method=self._me_endpoint.method,
            )
        ).get_success_or_raise()
        return TypeAdapter(Principal).validate_json(response.body)

    def retrieve(self, items: Sequence[PrincipalId | ExternalId], ignore_unknown_ids: bool = False) -> list[Principal]:
        """Retrieve principals by their IDs or external IDs.

        Args:
            items: A sequence of Principal or ExternalId objects to retrieve.
            ignore_unknown_ids: If True, unknown IDs will be ignored and not cause an error. Default is False.
        """
        path = self._method_endpoint_map["retrieve"].path.format(org=self._project_api.get_organization_id())
        return self._request_item_response(
            items, "retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}, endpoint=path
        )

    def _create_list_parameters(self, types: list[PrincipalType] | None) -> dict[str, PrimitiveType] | None:
        """Create the query parameters for the list endpoint."""
        if types is None:
            return None
        return {"types": ",".join(types)}

    def paginate(
        self, types: list[PrincipalType] | None = None, limit: int = 10, cursor: str | None = None
    ) -> PagedResponse[Principal]:
        """Paginate through all principals in the organization.

        Args:
            limit: The maximum number of items to return per page. Default is 10.

        Returns:
            A sequence of Principal objects.
        """
        path = self._method_endpoint_map["list"].path.format(org=self._project_api.get_organization_id())
        return self._paginate(limit, cursor, params=self._create_list_parameters(types), endpoint_path=path)

    def iterate(
        self, types: list[PrincipalType] | None = None, limit: int | None = 10, cursor: str | None = None
    ) -> Iterable[list[Principal]]:
        """Iterate through all principals in the organization."""
        endpoint_path = self._method_endpoint_map["list"].path.format(org=self._project_api.get_organization_id())
        return self._iterate(
            params=self._create_list_parameters(types), endpoint_path=endpoint_path, limit=limit, cursor=cursor
        )

    def list(self, types: list[PrincipalType] | None = None, limit: int = 10) -> list[Principal]:
        """List principals in the organization.

        Args:
            limit: The maximum number of items to return. Default is 10.

        Returns:
            A list of Principal objects.
        """
        endpoint_path = self._method_endpoint_map["list"].path.format(org=self._project_api.get_organization_id())
        return self._list(params=self._create_list_parameters(types), endpoint_path=endpoint_path, limit=limit)



from typing import Any


class ToolkitAPIError(Exception):
    """Base class for all exceptions raised by the Cognite Toolkit API client."""

    def __init__(
        self,
        message: str,
        missing: list[dict[str, Any]] | None = None,
        duplicated: list[dict[str, Any]] | None = None,
        code: int | None = None,
        is_auto_retryable: bool | None = None,
    ) -> None:
        super().__init__(message)
        self.message = message
        self.missing = missing
        self.duplicated = duplicated
        self.code = code
        self.is_auto_retryable = is_auto_retryable



from typing import ClassVar

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from ._common import Transform


class RobotFrame(BaseModelObject):
    """The frames resource represents coordinate frames, which are used to describe how maps are aligned with
    respect to each other. For example, frames are used to describe the relative position of a context map
    (e.g., a 3D model of a location) and a robot's navigation map. Frames are aligned with each other through
    transforms, which consist of a translation (in meters) and rotation (quaternion).

    Args:
        name: Frame name.
        external_id: Frame external id. Must be unique for the resource type.
        transform: Transform of the parent frame to the current frame.
    """

    external_id: str
    name: str
    transform: Transform | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RobotFrameRequest(RobotFrame, UpdatableRequestResource):
    """Request resource for creating or updating a Frame."""

    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"transform"})


class RobotFrameResponse(RobotFrame, ResponseResource[RobotFrameRequest]):
    """Response resource for a Frame."""

    created_time: int
    updated_time: int

    @classmethod
    def request_cls(cls) -> type[RobotFrameRequest]:
        return RobotFrameRequest



import collections.abc
import io
from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, final

import pandas as pd
from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter, SequenceRowFilter
from cognite_toolkit._cdf_tk.client.resource_classes.asset import (
    AssetAggregateItem,
    AssetRequest,
    AssetResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.event import EventRequest, EventResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.sequence import SequenceRequest, SequenceResponse
from cognite_toolkit._cdf_tk.client.resource_classes.sequence_rows import SequenceRowsRequest, SequenceRowsResponse
from cognite_toolkit._cdf_tk.constants import TABLE_FORMATS, YAML_SUFFIX
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.resource_classes import AssetYAML, EventYAML, SequenceRowYAML, SequenceYAML
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning, ToolkitDeprecationWarning
from cognite_toolkit._cdf_tk.utils import load_yaml_inject_variables
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable
from cognite_toolkit._cdf_tk.utils.file import read_csv

from .data_organization import DataSetsCRUD, LabelCRUD

_DEPRECATION_WARNING_ISSUED = False


@final
class AssetCRUD(ResourceCRUD[ExternalId, AssetRequest, AssetResponse]):
    folder_name = "classic"
    resource_cls = AssetResponse
    resource_write_cls = AssetRequest
    yaml_cls = AssetYAML
    kind = "Asset"
    dependencies = frozenset({DataSetsCRUD, LabelCRUD})
    _doc_url = "Assets/operation/createAssets"

    @classmethod
    def is_supported_file(cls, file: Path) -> bool:
        if Flags.v08.is_enabled():
            return super().is_supported_file(file)
        global _DEPRECATION_WARNING_ISSUED
        if not file.stem.casefold().endswith(cls.kind.casefold()):
            return False
        if file.suffix in YAML_SUFFIX:
            return True
        if file.suffix in TABLE_FORMATS:
            if not _DEPRECATION_WARNING_ISSUED:
                ToolkitDeprecationWarning(
                    feature="deployment of asset from CSV or Parquet files",
                    alternative="data plugin and cdf data upload commands",
                    removal_version="0.8",
                ).print_warning()
                _DEPRECATION_WARNING_ISSUED = True
            return True
        return False

    @property
    def display_name(self) -> str:
        return "assets"

    @classmethod
    def get_id(cls, item: AssetRequest | AssetResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("Asset must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: AssetResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Asset must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[AssetRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.AssetsAcl.Scope.All | capabilities.AssetsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.AssetsAcl.Scope.All()
        )

        actions = (
            [capabilities.AssetsAcl.Action.Read]
            if read_only
            else [capabilities.AssetsAcl.Action.Read, capabilities.AssetsAcl.Action.Write]
        )

        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.AssetsAcl.Scope.DataSet(list(data_set_ids))

        return capabilities.AssetsAcl(actions, scope)

    def create(self, items: collections.abc.Sequence[AssetRequest]) -> list[AssetResponse]:
        return self.client.tool.assets.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[AssetResponse]:
        return self.client.tool.assets.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: collections.abc.Sequence[AssetRequest]) -> list[AssetResponse]:
        return self.client.tool.assets.update(items, mode="replace")

    def delete(self, ids: Sequence[InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.assets.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[AssetResponse]:
        filter_ = ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=data_set_external_id)
        for assets in self.client.tool.assets.iterate(aggregated_properties=True, filter=filter_, limit=None):
            yield from assets

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        for label in item.get("labels", []):
            if isinstance(label, dict):
                yield LabelCRUD, ExternalId(external_id=label["externalId"])
            elif isinstance(label, str):
                yield LabelCRUD, ExternalId(external_id=label)
        if "parentExternalId" in item:
            yield cls, item["parentExternalId"]

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources: list[dict[str, Any]]
        if filepath.suffix in {".yaml", ".yml"}:
            raw_yaml = load_yaml_inject_variables(
                self.safe_read(filepath),
                environment_variables or {},
                original_filepath=filepath,
            )
            resources = [raw_yaml] if isinstance(raw_yaml, dict) else raw_yaml
        elif filepath.suffix == ".csv" or filepath.suffix == ".parquet":
            if filepath.suffix == ".csv":
                # The replacement is used to ensure that we read exactly the same file on Windows and Linux
                file_content = filepath.read_bytes().replace(b"\r\n", b"\n").decode("utf-8")
                data = read_csv(io.StringIO(file_content))
            else:
                data = pd.read_parquet(filepath)
            data.replace(pd.NA, None, inplace=True)
            data.replace("", None, inplace=True)
            resources = data.to_dict(orient="records")
        else:
            raise ValueError(f"Unsupported file type: {filepath.suffix}")

        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> AssetRequest:
        # Unpack metadata keys from table formats (e.g. csv, parquet)
        metadata: dict = resource.get("metadata", {})
        for key, value in list(resource.items()):
            if key.startswith("metadata."):
                if value not in {None, float("nan")} and str(value) not in {"", " ", "nan", "null", "none"}:
                    metadata[key.removeprefix("metadata.")] = str(value)
                del resource[key]
        if metadata:
            resource["metadata"] = metadata
        if isinstance(resource.get("labels"), str):
            resource["labels"] = [
                label.strip() for label in resource["labels"].removeprefix("[").removesuffix("]").split(",")
            ]

        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return AssetRequest.model_validate(resource)

    def dump_resource(self, resource: AssetResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        if "parentId" in dumped and "parentId" not in local:
            dumped.pop("parentId")
        if resource.aggregates:
            # This is only included when the asset is downloaded/migrated with aggregated properties
            aggregates = (
                resource.aggregates.dump()
                if isinstance(resource.aggregates, AssetAggregateItem)
                else resource.aggregates
            )
            if "path" in aggregates:
                path = aggregates.pop("path", [])
                if path:
                    aggregates["path"] = self.client.lookup.assets.external_id(
                        [segment["id"] for segment in path if "id" in segment]
                    )
            dumped.update(aggregates)
        return dumped


@final
class SequenceCRUD(ResourceCRUD[ExternalId, SequenceRequest, SequenceResponse]):
    folder_name = "classic"
    resource_cls = SequenceResponse
    resource_write_cls = SequenceRequest
    kind = "Sequence"
    dependencies = frozenset({DataSetsCRUD, AssetCRUD})
    yaml_cls = SequenceYAML
    _doc_url = "Sequences/operation/createSequence"

    @property
    def display_name(self) -> str:
        return "sequences"

    @classmethod
    def get_id(cls, item: SequenceRequest | SequenceResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("Sequence must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SequenceResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Sequence must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[SequenceRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: Any = capabilities.SequencesAcl.Scope.All()
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.SequencesAcl.Scope.DataSet(list(data_set_ids))

        actions = (
            [capabilities.SequencesAcl.Action.Read]
            if read_only
            else [capabilities.SequencesAcl.Action.Read, capabilities.SequencesAcl.Action.Write]
        )

        return capabilities.SequencesAcl(actions, scope)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SequenceRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if asset_external_id := resource.pop("assetExternalId", None):
            resource["assetId"] = self.client.lookup.assets.id(asset_external_id, is_dry_run)
        return SequenceRequest.model_validate(resource)

    def dump_resource(self, resource: SequenceResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if asset_id := dumped.pop("assetId", None):
            dumped["assetExternalId"] = self.client.lookup.assets.external_id(asset_id)
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        local_col_by_id = {col["externalId"]: col for col in local.get("columns", []) if "externalId" in col}
        for col in dumped.get("columns", []):
            external_id = col.get("externalId")
            if not external_id:
                continue
            if external_id not in local_col_by_id:
                continue
            local_col = local_col_by_id[external_id]
            if not col.get("metadata") and "metadata" not in local_col:
                col.pop("metadata", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path != ("columns",):
            return super().diff_list(local, cdf, json_path)
        return diff_list_identifiable(local, cdf, get_identifier=lambda col: col["externalId"])

    def create(self, items: collections.abc.Sequence[SequenceRequest]) -> list[SequenceResponse]:
        return self.client.tool.sequences.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SequenceResponse]:
        return self.client.tool.sequences.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: collections.abc.Sequence[SequenceRequest]) -> list[SequenceResponse]:
        return self.client.tool.sequences.update(items, mode="replace")

    def delete(self, ids: Sequence[InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.sequences.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SequenceResponse]:
        filter_ = ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=data_set_external_id)
        for sequences in self.client.tool.sequences.iterate(filter=filter_):
            yield from sequences

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        if "assetExternalId" in item:
            yield AssetCRUD, ExternalId(external_id=item["assetExternalId"])


@final
class SequenceRowCRUD(ResourceCRUD[ExternalId, SequenceRowsRequest, SequenceRowsResponse]):
    folder_name = "classic"
    resource_cls = SequenceRowsResponse
    resource_write_cls = SequenceRowsRequest
    kind = "SequenceRow"
    dependencies = frozenset({SequenceCRUD})
    parent_resource = frozenset({SequenceCRUD})
    _doc_url = "Sequences/operation/postSequenceData"
    yaml_cls = SequenceRowYAML
    support_update = False

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console)
        # Used in the .diff_list method to keep track of the last column in the local list
        # such that the values in the rows can be matched to the correct column.
        self._last_column: tuple[dict[int, int], list[int]] = {}, []

    @property
    def display_name(self) -> str:
        return "sequence rows"

    @classmethod
    def get_id(cls, item: SequenceRowsRequest | SequenceRowsResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("SequenceRows must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SequenceRowsResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SequenceRows must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[SequenceRowsRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # We don't have any capabilities for SequenceRows, that is already handled by the Sequence
        return []

    def create(self, items: Sequence[SequenceRowsRequest]) -> Sequence[SequenceRowsRequest]:
        self.client.tool.sequences.rows.create(list(items))
        return items

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SequenceRowsResponse]:
        results: list[SequenceRowsResponse] = []
        for id_ in ids:
            row_filter = SequenceRowFilter(external_id=id_.external_id)
            responses = self.client.tool.sequences.rows.list(row_filter)
            results.extend(responses)
        return results

    def delete(self, ids: Sequence[ExternalId]) -> int:
        deleted: int = 0
        for id_ in ids:
            row_filter = SequenceRowFilter(external_id=id_.external_id)
            for batch in self.client.tool.sequences.rows.iterate(row_filter, limit=None):
                if not batch or not batch[0].rows:
                    continue
                item = batch[0]
                self.client.tool.sequences.rows.delete([item.as_request_resource().as_id()])
                deleted += len(item.rows)
        return deleted

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SequenceRowsResponse]:
        if parent_ids is None:
            filter_ = ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=data_set_external_id)
            parent_external_ids: list[str] = []
            for sequences in self.client.tool.sequences.iterate(filter=filter_):
                parent_external_ids.extend(seq.external_id for seq in sequences if seq.external_id)
        else:
            parent_external_ids = [id.external_id for id in parent_ids if isinstance(id, ExternalId)]
        for ext_id in parent_external_ids:
            row_filter = SequenceRowFilter(external_id=ext_id)
            responses = self.client.tool.sequences.rows.list(row_filter)
            yield from responses

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        yield SequenceCRUD, ExternalId(external_id=item["externalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SequenceRowsRequest:
        return SequenceRowsRequest.model_validate(resource)

    def dump_resource(self, resource: SequenceRowsResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        # Ensure that the rows is the last key in the dumped dictionary,
        # This information is used in the .diff_list method to match the values in the rows to the correct column.
        dumped["rows"] = dumped.pop("rows")
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("rows",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda row: row["rowNumber"])
        elif json_path == ("columns",):
            self._last_column = diff_list_hashable(local, cdf)
            # This is a special case where we need to keep track of the last column in the local list
            # such that the values in the rows can be matched to the correct column.
            return self._last_column
        elif len(json_path) == 3 and json_path[0] == "rows" and json_path[2] == "values":
            local_by_cdf, added = self._last_column
            if len(cdf) == len(local_by_cdf) + len(added):
                return local_by_cdf, added
            else:
                LowSeverityWarning("Number of rows in does not match the number of columns").print_warning()
                # Just assume that the rows are in the correct order
                return {no: no for no in range(len(cdf))}, []
        return super().diff_list(local, cdf, json_path)


@final
class EventCRUD(ResourceCRUD[ExternalId, EventRequest, EventResponse]):
    folder_name = "classic"
    resource_cls = EventResponse
    resource_write_cls = EventRequest
    yaml_cls = EventYAML
    kind = "Event"
    dependencies = frozenset({DataSetsCRUD, AssetCRUD})
    _doc_url = "Events/operation/createEvents"

    @property
    def display_name(self) -> str:
        return "events"

    @classmethod
    def get_id(cls, item: EventRequest | EventResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("Event must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: EventResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Event must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[EventRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.EventsAcl.Scope.All | capabilities.EventsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.EventsAcl.Scope.All()
        )

        actions = (
            [capabilities.EventsAcl.Action.Read]
            if read_only
            else [capabilities.EventsAcl.Action.Read, capabilities.EventsAcl.Action.Write]
        )

        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.EventsAcl.Scope.DataSet(list(data_set_ids))

        return capabilities.EventsAcl(actions, scope)

    def create(self, items: collections.abc.Sequence[EventRequest]) -> list[EventResponse]:
        return self.client.tool.events.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[EventResponse]:
        return self.client.tool.events.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: collections.abc.Sequence[EventRequest]) -> list[EventResponse]:
        return self.client.tool.events.update(items, mode="replace")

    def delete(self, ids: Sequence[InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.events.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[EventResponse]:
        filter_ = ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=data_set_external_id)
        for events in self.client.tool.events.iterate(filter=filter_, limit=None):
            yield from events

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        for asset_id in item.get("assetExternalIds", []):
            if isinstance(asset_id, str):
                yield AssetCRUD, ExternalId(external_id=asset_id)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> EventRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if asset_external_ids := resource.pop("assetExternalIds", []):
            resource["assetIds"] = self.client.lookup.assets.id(asset_external_ids, is_dry_run)
        return EventRequest.model_validate(resource)

    def dump_resource(self, resource: EventResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if asset_ids := dumped.pop("assetIds", None):
            local_order = {asset: no for no, asset in enumerate(local.get("assetExternalIds", []))}
            end_of_list = len(local_order)
            dumped["assetExternalIds"] = sorted(
                self.client.lookup.assets.external_id(asset_ids), key=lambda a: local_order.get(a, end_of_list)
            )
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path != ("assetExternalIds",):
            return super().diff_list(local, cdf, json_path)
        return diff_list_hashable(local, cdf)



from collections.abc import Callable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import get_args

from rich import print
from rich.console import Console
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsFailedRequest,
    ItemsFailedResponse,
    ItemsSuccessResponse,
)
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands._migrate.creators import MigrationCreator
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import DataMapper
from cognite_toolkit._cdf_tk.commands.deploy import DeployCommand
from cognite_toolkit._cdf_tk.constants import DMS_INSTANCE_LIMIT_MARGIN
from cognite_toolkit._cdf_tk.cruds import ResourceWorker
from cognite_toolkit._cdf_tk.data_classes import DeployResults
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitFileExistsError,
    ToolkitMigrationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.protocols import T_ResourceRequest, T_ResourceResponse
from cognite_toolkit._cdf_tk.storageio import T_Selector, UploadableStorageIO, UploadItem
from cognite_toolkit._cdf_tk.storageio.logger import FileDataLogger, OperationStatus
from cognite_toolkit._cdf_tk.utils import humanize_collection, safe_write, sanitize_filename
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.fileio import NDJsonWriter, Uncompressed
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor

from .data_model import INSTANCE_SOURCE_VIEW_ID, MODEL_ID, RESOURCE_VIEW_MAPPING_VIEW_ID
from .issues import WriteIssue


@dataclass
class OperationIssue:
    message: str
    count: int


@dataclass
class MigrationStatusResult:
    status: OperationStatus
    issues: list[OperationIssue]
    count: int


class MigrationCommand(ToolkitCommand):
    def migrate(
        self,
        selected: T_Selector,
        data: UploadableStorageIO[T_Selector, T_ResourceResponse, T_ResourceRequest],
        mapper: DataMapper[T_Selector, T_ResourceResponse, T_ResourceRequest],
        log_dir: Path,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> list[MigrationStatusResult]:
        if log_dir.exists() and any(log_dir.iterdir()):
            raise ToolkitFileExistsError(
                f"Log directory {log_dir} already exists. Please remove it or choose another directory."
            )
        self.validate_migration_model_available(data.client)
        log_dir.mkdir(parents=True, exist_ok=True)
        mapper.prepare(selected)

        iteration_count: int | None = None
        total_items = data.count(selected)
        if total_items is not None:
            iteration_count = (total_items // data.CHUNK_SIZE) + (1 if total_items % data.CHUNK_SIZE > 0 else 0)
            self.validate_available_capacity(data.client, total_items)

        console = Console()
        with (
            NDJsonWriter(log_dir, kind=f"{selected.kind}MigrationIssues", compression=Uncompressed) as log_file,
            HTTPClient(config=data.client.config) as write_client,
        ):
            logger = FileDataLogger(log_file)
            data.logger = logger
            mapper.logger = logger

            executor = ProducerWorkerExecutor[Sequence[T_ResourceResponse], Sequence[UploadItem[T_ResourceRequest]]](
                download_iterable=(page.items for page in data.stream_data(selected)),
                process=self._convert(mapper, data),
                write=self._upload(selected, write_client, data, dry_run),
                iteration_count=iteration_count,
                max_queue_size=10,
                download_description=f"Downloading {selected.display_name}",
                process_description="Converting",
                write_description="Uploading",
                console=console,
                verbose=verbose,
            )

            executor.run()
            total = executor.total_items

        results = self._create_status_summary(logger)

        self._print_rich_tables(results, console)
        self._print_txt(results, log_dir, f"{selected.kind}Items", console)
        executor.raise_on_error()
        action = "Would migrate" if dry_run else "Migrating"
        console.print(f"{action} {total:,} {selected.display_name} to instances.")

        return results

    # Todo: Move to the logger module
    @classmethod
    def _create_status_summary(cls, logger: FileDataLogger) -> list[MigrationStatusResult]:
        results: list[MigrationStatusResult] = []
        status_counts = logger.tracker.get_status_counts()
        for status in get_args(OperationStatus):
            issue_counts = logger.tracker.get_issue_counts(status)
            issues = [OperationIssue(message=issue, count=count) for issue, count in issue_counts.items()]
            result = MigrationStatusResult(
                status=status,
                issues=issues,
                count=status_counts.get(status, 0),
            )
            results.append(result)
        return results

    def _print_rich_tables(self, results: list[MigrationStatusResult], console: Console) -> None:
        table = Table(title="Migration Summary", show_lines=True)
        table.add_column("Status", style="bold")
        table.add_column("Count", justify="right", style="bold")
        table.add_column("Issues", style="bold")
        for result in results:
            issues_str = "\n".join(f"{issue.message}: {issue.count}" for issue in result.issues) or ""
            table.add_row(result.status, str(result.count), issues_str)
        console.print(table)

    def _print_txt(self, results: list[MigrationStatusResult], log_dir: Path, kind: str, console: Console) -> None:
        summary_file = log_dir / f"{kind}_migration_summary.txt"
        with summary_file.open("w", encoding="utf-8") as f:
            f.write("Migration Summary\n")
            f.write("=================\n\n")
            for result in results:
                f.write(f"Status: {result.status}\n")
                f.write(f"Count: {result.count}\n")
                f.write("Issues:\n")
                if result.issues:
                    for issue in result.issues:
                        f.write(f"  - {issue.message}: {issue.count}\n")
                else:
                    f.write("  None\n")
                f.write("\n")
        console.print(f"Summary written to {log_dir}")

    @staticmethod
    def _convert(
        mapper: DataMapper[T_Selector, T_ResourceResponse, T_ResourceRequest],
        data: UploadableStorageIO[T_Selector, T_ResourceResponse, T_ResourceRequest],
    ) -> Callable[[Sequence[T_ResourceResponse]], Sequence[UploadItem[T_ResourceRequest]]]:
        def track_mapping(source: Sequence[T_ResourceResponse]) -> list[UploadItem[T_ResourceRequest]]:
            mapped = mapper.map(source)
            return [
                UploadItem(source_id=data.as_id(item), item=target)
                for target, item in zip(mapped, source)
                if target is not None
            ]

        return track_mapping

    def _upload(
        self,
        selected: T_Selector,
        write_client: HTTPClient,
        target: UploadableStorageIO[T_Selector, T_ResourceResponse, T_ResourceRequest],
        dry_run: bool,
    ) -> Callable[[Sequence[UploadItem[T_ResourceRequest]]], None]:
        def upload_items(data_item: Sequence[UploadItem[T_ResourceRequest]]) -> None:
            if not data_item:
                return None
            if dry_run:
                target.logger.tracker.finalize_item([item.source_id for item in data_item], "pending")
                return None

            responses = target.upload_items(data_chunk=data_item, http_client=write_client, selector=selected)

            # Todo: Move logging into the UploadableStorageIO class
            issues: list[WriteIssue] = []
            for item in responses:
                if isinstance(item, ItemsSuccessResponse):
                    target.logger.tracker.finalize_item(item.ids, "success")
                    continue
                if isinstance(item, ItemsFailedResponse):
                    error = item.error
                    for id_ in item.ids:
                        issue = WriteIssue(id=str(id_), status_code=error.code, message=error.message)
                        issues.append(issue)
                elif isinstance(item, ItemsFailedRequest):
                    for id_ in item.ids:
                        issue = WriteIssue(id=str(id_), status_code=0, message=item.error_message)
                        issues.append(issue)

                if isinstance(item, ItemsFailedResponse | ItemsFailedRequest):
                    target.logger.tracker.finalize_item(item.ids, "failure")
            if issues:
                target.logger.log(issues)
            return None

        return upload_items

    @staticmethod
    def validate_migration_model_available(client: ToolkitClient) -> None:
        models = client.tool.data_models.retrieve([MODEL_ID], inline_views=False)
        if not models:
            raise ToolkitMigrationError(
                f"The migration data model {MODEL_ID!r} does not exist. "
                "Please run the `cdf migrate prepare` command to deploy the migration data model."
            )
        elif len(models) > 1:
            raise ToolkitMigrationError(
                f"Multiple migration models {MODEL_ID!r}. "
                "Please delete the duplicate models before proceeding with the migration."
            )
        model = models[0]
        missing_views = {INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID} - set(model.views or [])
        if missing_views:
            raise ToolkitMigrationError(
                f"Invalid migration model. Missing views {humanize_collection(missing_views)}. "
                f"Please run the `cdf migrate prepare` command to deploy the migration data model."
            )

    def validate_available_capacity(self, client: ToolkitClient, instance_count: int) -> None:
        """Validate that the project has enough capacity to accommodate the migration."""

        stats = client.data_modeling.statistics.project()

        available_capacity = stats.instances.instances_limit - stats.instances.instances
        available_capacity_after = available_capacity - instance_count

        if available_capacity_after < DMS_INSTANCE_LIMIT_MARGIN:
            raise ToolkitValueError(
                "Cannot proceed with migration, not enough instance capacity available. Total capacity after migration"
                f" would be {available_capacity_after:,} instances, which is less than the required margin of"
                f" {DMS_INSTANCE_LIMIT_MARGIN:,} instances. Please increase the instance capacity in your CDF project"
                f" or delete some existing instances before proceeding with the migration of {instance_count:,} assets."
            )
        total_instances = stats.instances.instances + instance_count
        self.console(
            f"Project has enough capacity for migration. Total instances after migration: {total_instances:,}."
        )

    def create(
        self,
        client: ToolkitClient,
        creator: MigrationCreator,
        dry_run: bool,
        output_dir: Path,
        deploy: bool = True,
        verbose: bool = False,
    ) -> DeployResults:
        """This method is used to create migration resource in CDF."""
        self.validate_migration_model_available(client)

        deploy_cmd = DeployCommand(self.print_warning, silent=self.silent)
        deploy_cmd.tracker = self.tracker
        # crud_cls = creator.CRUD
        # resource_list = creator.create_resources()
        results = DeployResults([], "deploy", dry_run=dry_run)
        for to_create in creator.create_resources():
            crud_cls = to_create.crud_cls
            if deploy:
                crud = crud_cls.create_loader(client)
                worker = ResourceWorker(crud, "deploy")
                local_by_id = {
                    crud.get_id(item.resource): (item.resource.dump(), item.resource) for item in to_create.resources
                }
                worker.validate_access(local_by_id, is_dry_run=dry_run)
                cdf_resources = crud.retrieve(list(local_by_id.keys()))
                resources = worker.categorize_resources(local_by_id, cdf_resources, False, verbose)

                if dry_run:
                    result = deploy_cmd.dry_run_deploy(resources, crud, False, False)
                else:
                    result = deploy_cmd.actual_deploy(resources, crud)
                    if result.calculated_total > 0 and to_create.store_linage is not None:
                        store_count = to_create.store_linage()
                        self.console(f"Stored lineage for {store_count:,} {to_create.display_name}.")

                verb = "Would deploy" if dry_run else "Deploying"
                self.console(f"{verb} {to_create.display_name} to CDF.")

                if result:
                    results[result.name] = result

                if results.has_counts:
                    print(results.counts_table())

            for item in to_create.resources:
                if item.config_data and item.filestem:
                    filepath = (
                        output_dir / crud_cls.folder_name / f"{sanitize_filename(item.filestem)}.{crud_cls.kind}.yaml"
                    )
                    filepath.parent.mkdir(parents=True, exist_ok=True)
                    safe_write(filepath, yaml_safe_dump(item.config_data))
            self.console(
                f"{len(to_create.resources)} {crud_cls.kind} resource configurations written to {(output_dir / crud_cls.folder_name).as_posix()!r}"
            )
            self.console("It is recommended to add these files to a Toolkit governed module.")

        return results



from cognite_toolkit._cdf_tk.client.http_client import HTTPClient

from .simulator_model_revisions import SimulatorModelRevisionsAPI
from .simulator_models import SimulatorModelsAPI
from .simulator_routine_revisions import SimulatorRoutineRevisionsAPI
from .simulator_routines import SimulatorRoutinesAPI


class SimulatorsAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self.models = SimulatorModelsAPI(http_client)
        self.model_revisions = SimulatorModelRevisionsAPI(http_client)
        self.routines = SimulatorRoutinesAPI(http_client)
        self.routine_revisions = SimulatorRoutineRevisionsAPI(http_client)



from datetime import date
from pathlib import Path
from typing import Annotated, Any

import questionary
import typer

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import ContainerReference
from cognite_toolkit._cdf_tk.commands import MigrationPrepareCommand
from cognite_toolkit._cdf_tk.commands._migrate import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.creators import (
    InfieldV2ConfigCreator,
    InstanceSpaceCreator,
    SourceSystemCreator,
)
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import (
    AssetCentricMapper,
    CanvasMapper,
    ChartMapper,
    ThreeDAssetMapper,
    ThreeDMapper,
)
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import (
    AnnotationMigrationIO,
    AssetCentricMigrationIO,
    ThreeDAssetMappingMigrationIO,
    ThreeDMigrationIO,
)
from cognite_toolkit._cdf_tk.commands._migrate.selectors import (
    AssetCentricMigrationSelector,
    MigrateDataSetSelector,
    MigrationCSVFileSelector,
)
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.storageio import CanvasIO, ChartIO
from cognite_toolkit._cdf_tk.storageio.selectors import (
    CanvasExternalIdSelector,
    ChartExternalIdSelector,
    ThreeDModelIdSelector,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.cli_args import parse_view_str
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    APMConfigInteractiveSelect,
    AssetInteractiveSelect,
    DataModelingSelect,
    FileMetadataInteractiveSelect,
    InteractiveCanvasSelect,
    InteractiveChartSelect,
    ResourceViewMappingInteractiveSelect,
    ThreeDInteractiveSelect,
    ViewSelectFilter,
)
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricKind

TODAY = date.today()


class MigrateApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("prepare")(self.prepare)
        self.command("data-sets")(self.data_sets)
        self.command("source-systems")(self.source_systems)
        self.command("assets")(self.assets)
        self.command("events")(self.events)
        self.command("timeseries")(self.timeseries)
        self.command("files")(self.files)
        self.command("annotations")(self.annotations)
        self.command("canvas")(self.canvas)
        self.command("charts")(self.charts)
        self.command("3d")(self.three_d)
        self.command("3d-mappings")(self.three_d_asset_mapping)
        if Flags.INFIELD_MIGRATE.is_enabled():
            self.command("infield-configs")(self.infield_configs)

    def main(self, ctx: typer.Context) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf migrate --help[/] for more information.")

    @staticmethod
    def prepare(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="The preparation will not be executed, only report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Prepare the migration of resources from Asset-Centric to data modeling in CDF.

        This means deploying the CogniteMigration data model that contains the Mapping view. This will be used
        to store the mapping from Asset-Centric resources to the new data modeling resources.

        This mapping will be used when migrating applications such as Canvas, Charts, as well as resources that
        depend on the primary resources 3D and annotations.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = MigrationPrepareCommand(client=client)
        cmd.run(
            lambda: cmd.deploy_cognite_migration(
                client,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def data_sets(
        ctx: typer.Context,
        data_set: Annotated[
            list[str] | None,
            typer.Argument(
                help="The name or external ID of the data set to create Instance Spaces for. If not provided, an "
                "interactive selection will be performed to select the data sets to create Instance Spaces for."
            ),
        ] = None,
        auto_fix: Annotated[
            bool,
            typer.Option(
                "--auto-fix",
                "-a",
                help="If set, the command will try to make the data set external ID a valid space identifier,"
                "if it is not already.",
            ),
        ] = False,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the instance space definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates Instance Spaces for all selected data sets."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_set is None:
            # Interactive model
            selector = AssetInteractiveSelect(client, "migrate")
            data_set = selector.select_data_sets()
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            output_dir = Path(
                questionary.path(
                    "Specify output directory for instance space definitions:", default=str(output_dir)
                ).unsafe_ask()
            )
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.create(
                client,
                creator=InstanceSpaceCreator(client, data_set_external_ids=data_set, auto_fix=auto_fix),
                output_dir=output_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def source_systems(
        ctx: typer.Context,
        data_set: Annotated[
            str | None,
            typer.Argument(
                help="The external ID of the data set to lookup source system for. If not provided, an interactive "
                "selection will be performed to select the data sets to create instance spaces for."
            ),
        ] = None,
        instance_space: Annotated[
            str | None,
            typer.Option(
                "--instance-space",
                "-s",
                help="The instance space were you want to create the source system.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the instance space definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates source systems from the 'source' property of classic resources (assets, events, files)."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_set is None and instance_space is None:
            # Interactive model
            ds_selector = AssetInteractiveSelect(client, "migrate")
            data_set = ds_selector.select_data_set()
            dm_selector = DataModelingSelect(client, "migrate")
            instance_space = dm_selector.select_instance_space(
                multiselect=False,
                message="In which instance space do you want to create the source system?",
                include_empty=True,
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            output_dir = Path(
                questionary.path(
                    "Specify output directory for instance space definitions:", default=str(output_dir)
                ).unsafe_ask()
            )
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()
        elif data_set is None or instance_space is None:
            raise typer.BadParameter("Both data_set and instance_space must be provided together.")

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.create(
                client,
                creator=SourceSystemCreator(client, data_set_external_id=data_set, instance_space=instance_space),
                output_dir=output_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def assets(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Assets to CogniteAssets. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to use for the migrated CogniteAssets. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated assets. If not provided, "
                "the default mapping to CogniteAsset in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated assets Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating assets. If not provided, "
                "CogniteAsset in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="(Only used with mapping-file) If set, no confirmation prompt will be shown before proceeding with the migration.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Assets to CogniteAssets."""
        client = EnvironmentVariables.create_from_environment().get_client()
        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            auto_yes=auto_yes,
            verbose=verbose,
            kind="Assets",
            resource_type="asset",
            container_id=ContainerReference(space="cdf_cdm", external_id="CogniteAsset"),
        )

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def _prepare_asset_centric_arguments(
        client: ToolkitClient,
        mapping_file: Path | None,
        data_set_id: str | None,
        consumption_view: str | None,
        ingestion_mapping: str | None,
        auto_yes: bool,
        dry_run: bool,
        verbose: bool,
        kind: AssetCentricKind,
        resource_type: str,
        container_id: ContainerReference,
    ) -> tuple[AssetCentricMigrationSelector, bool, bool]:
        if data_set_id is not None and mapping_file is not None:
            raise typer.BadParameter("Cannot specify both data_set_id and mapping_file")
        elif mapping_file is not None:
            file_selector = MigrationCSVFileSelector(datafile=mapping_file, kind=kind)
            selected: AssetCentricMigrationSelector = file_selector

            panel = file_selector.items.print_status()
            if panel is not None:
                client.console.print(panel)
                if not auto_yes:
                    proceed = questionary.confirm(
                        "Do you want to proceed with the migration?", default=False
                    ).unsafe_ask()
                    if not proceed:
                        client.console.print("Migration aborted by user.")
                        raise typer.Abort()
        elif data_set_id is not None:
            parsed_view = parse_view_str(consumption_view) if consumption_view is not None else None
            selected = MigrateDataSetSelector(
                data_set_external_id=data_set_id,
                kind=kind,
                ingestion_mapping=ingestion_mapping,
                preferred_consumer_view=parsed_view,
            )
        else:
            # Interactive selection of data set.
            selector = AssetInteractiveSelect(client, "migrate")
            selected_data_set_id = selector.select_data_set(allow_empty=False)
            asset_mapping = ResourceViewMappingInteractiveSelect(client, "migrate").select_resource_view_mapping(
                resource_type,
            )
            preferred_consumer_view = (
                DataModelingSelect(client, "migrate")
                .select_view(
                    multiselect=False,
                    filter=ViewSelectFilter(
                        include_global=True,
                        mapped_container=container_id,
                        instance_type="node",
                    ),
                )
                .as_id()
            )
            selected = MigrateDataSetSelector(
                data_set_external_id=selected_data_set_id,
                kind=kind,
                ingestion_mapping=asset_mapping.external_id,
                preferred_consumer_view=preferred_consumer_view,
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()

        return selected, dry_run, verbose

    @classmethod
    def events(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Events to CogniteActivity. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to use for the migrated CogniteActivity. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated events. If not provided, "
                "the default mapping to CogniteActivity in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated events Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating events. If not provided, "
                "CogniteActivity in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="(Only used with mapping-file) If set, no confirmation prompt will be shown before proceeding with the migration.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Events to CogniteActivity."""
        client = EnvironmentVariables.create_from_environment().get_client()
        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            auto_yes=auto_yes,
            verbose=verbose,
            kind="Events",
            resource_type="event",
            container_id=ContainerReference(space="cdf_cdm", external_id="CogniteActivity"),
        )

        cmd = MigrationCommand(client=client)

        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def timeseries(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from TimeSeries to CogniteTimeSeries. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to select for the timeseries to migrate. If not provided and the mapping file is not provided"
                "an interactive selection will be performed to select the data set to migrate timeseries from.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated timeseries. If not provided, "
                "the default mapping to CogniteTimeSeries in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated timeseries Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating timeseries. If not provided, "
                "CogniteTimeSeries in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        skip_linking: Annotated[
            bool,
            typer.Option(
                "--skip-linking",
                "-x",
                help="If set, the migration will not create links between the old TimeSeries and the new CogniteTimeSeries.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="(Only used with mapping-file) If set, no confirmation prompt will be shown before proceeding with the migration.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate TimeSeries to CogniteTimeSeries."""
        client = EnvironmentVariables.create_from_environment().get_client()

        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            auto_yes=auto_yes,
            verbose=verbose,
            kind="TimeSeries",
            resource_type="timeseries",
            container_id=ContainerReference(space="cdf_cdm", external_id="CogniteTimeSeries"),
        )
        if data_set_id is None and mapping_file is None:
            skip_linking = not questionary.confirm(
                "Do you want to link old and new TimeSeries?", default=not skip_linking
            ).unsafe_ask()

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client, skip_linking=skip_linking),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def files(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Files to CogniteFiles. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to select for the files to migrate. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated files. If not provided, "
                "the default mapping to CogniteFile in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated files Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating timeseries. If not provided, "
                "CogniteFile in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        skip_linking: Annotated[
            bool,
            typer.Option(
                "--skip-linking",
                "-x",
                help="If set, the migration will not create links between the old FileMetadata and the new CogniteFile.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="(Only used with mapping-file) If set, no confirmation prompt will be shown before proceeding with the migration.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Files to CogniteFiles."""
        client = EnvironmentVariables.create_from_environment().get_client()

        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            auto_yes=auto_yes,
            verbose=verbose,
            kind="FileMetadata",
            resource_type="file",
            container_id=ContainerReference(space="cdf_cdm", external_id="CogniteFile"),
        )
        cmd = MigrationCommand(client=client)

        if data_set_id is None and mapping_file is None:
            skip_linking = not questionary.confirm(
                "Do you want to link old and new Files?", default=not skip_linking
            ).unsafe_ask()

        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client, skip_linking=skip_linking),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def annotations(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Annotations to CogniteDiagramAnnotation. "
                "This file is expected to have the following columns: [id, space, externalId, ingestionView].",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to select for the annotations to migrate. If not provided and the mapping file is not provided, "
                "an interactive selection will be performed to select the data set to migrate annotations from.",
            ),
        ] = None,
        instance_space: Annotated[
            str | None,
            typer.Option(
                "--instance-space",
                "-i",
                help="The instance space to use for the migrated annotations. Required when using --data-set-id.",
            ),
        ] = None,
        asset_annotation_mapping: Annotated[
            str | None,
            typer.Option(
                "--asset-annotation-mapping",
                "-a",
                help="The ingestion mapping to use for asset-linked annotations. If not provided, "
                "the default mapping (cdf_asset_annotations_mapping) will be used.",
            ),
        ] = None,
        file_annotation_mapping: Annotated[
            str | None,
            typer.Option(
                "--file-annotation-mapping",
                "-f",
                help="The ingestion mapping to use for file-linked annotations. If not provided, "
                "the default mapping (cdf_file_annotations_mapping) will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Annotations to CogniteDiagramAnnotation edges in data modeling.

        Annotations are diagram annotations that link assets or files to other resources. This command
        migrates them to edges in the data modeling space, preserving the relationships and metadata.
        """
        client = EnvironmentVariables.create_from_environment().get_client()

        if data_set_id is not None and mapping_file is not None:
            raise typer.BadParameter("Cannot specify both data_set_id and mapping_file")
        elif mapping_file is not None:
            selected: AssetCentricMigrationSelector = MigrationCSVFileSelector(
                datafile=mapping_file, kind="Annotations"
            )
            annotation_io = AnnotationMigrationIO(client)
        elif data_set_id is not None:
            if instance_space is None:
                raise typer.BadParameter("--instance-space is required when using --data-set-id")
            selected = MigrateDataSetSelector(data_set_external_id=data_set_id, kind="Annotations")
            annotation_io = AnnotationMigrationIO(
                client,
                instance_space=instance_space,
                default_asset_annotation_mapping=asset_annotation_mapping,
                default_file_annotation_mapping=file_annotation_mapping,
            )
        else:
            # Interactive selection
            selector = FileMetadataInteractiveSelect(client, "migrate")
            selected_data_set_id = selector.select_data_set(allow_empty=False)
            dm_selector = DataModelingSelect(client, "migrate")
            selected_instance_space = dm_selector.select_instance_space(
                multiselect=False,
                message="In which instance space do you want to create the annotations?",
                include_empty=True,
            )
            if selected_instance_space is None:
                raise typer.Abort()
            asset_annotations_selector = ResourceViewMappingInteractiveSelect(client, "migrate asset annotations")
            asset_annotation_mapping = asset_annotations_selector.select_resource_view_mapping(
                resource_type="assetAnnotation",
            ).external_id
            file_annotations_selector = ResourceViewMappingInteractiveSelect(client, "migrate file annotations")
            file_annotation_mapping = file_annotations_selector.select_resource_view_mapping(
                resource_type="fileAnnotation",
            ).external_id

            selected = MigrateDataSetSelector(data_set_external_id=selected_data_set_id, kind="Annotations")
            annotation_io = AnnotationMigrationIO(
                client,
                instance_space=selected_instance_space,
                default_asset_annotation_mapping=asset_annotation_mapping,
                default_file_annotation_mapping=file_annotation_mapping,
            )

            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=annotation_io,
                mapper=AssetCentricMapper[AnnotationResponse](client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def canvas(
        ctx: typer.Context,
        external_id: Annotated[
            list[str] | None,
            typer.Argument(
                help="The external ID of the Canvas to migrate. If not provided, and interactive selection will be "
                "performed to select the Canvas to migrate."
            ),
        ] = None,
        allow_missing_ref: Annotated[
            bool,
            typer.Option(
                "--allow-missing-ref",
                "-a",
                help="If set, the command will migrate Canvases that reference resources that have not been migrated to data modeling. "
                "If not set, the migration will fail if any referenced resource are missing.",
            ),
        ] = False,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where migration logs will be stored.",
            ),
        ] = Path(f"migration_logs_{TODAY}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of "
                "what would be done is printed. This is useful for checking that all resources referenced by the Canvas"
                "have been migrated to the new data modeling resources in CDF.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Canvas applications from Asset-Centric to data modeling in CDF.

        This command expects that the CogniteMigration data model is already deployed, and that the Mapping view
        is populated with the mapping from Asset-Centric resources to the new data modeling resources.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        if external_id is None:
            interactive = InteractiveCanvasSelect(client)
            external_id = interactive.select_external_ids()
            log_dir = Path(
                questionary.path("Specify log directory for migration logs:", default=str(log_dir)).unsafe_ask()
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()

        cmd = MigrationCommand(client=client)
        selector = CanvasExternalIdSelector(external_ids=tuple(external_id))
        cmd.run(
            lambda: cmd.migrate(
                selected=selector,
                data=CanvasIO(client, exclude_existing_version=True),
                mapper=CanvasMapper(client, dry_run=dry_run, skip_on_missing_ref=not allow_missing_ref),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def charts(
        ctx: typer.Context,
        external_id: Annotated[
            list[str] | None,
            typer.Argument(
                help="The external ID of the Chart to migrate. If not provided, an interactive selection will be "
                "performed to select the Charts to migrate."
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where migration logs will be stored.",
            ),
        ] = Path(f"migration_logs_{TODAY}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of "
                "what would be done is printed. This is useful for checking that all time series referenced by the Charts "
                "have been migrated to the new data modeling resources in CDF.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Charts from time series references to data modeling in CDF.

        This command expects that the CogniteMigration data model is already deployed, and that the Mapping view
        is populated with the mapping from time series to the new data modeling resources.
        """
        client = EnvironmentVariables.create_from_environment().get_client()

        selected_external_ids: list[str]
        if external_id:
            selected_external_ids = external_id
        else:
            selected_external_ids = InteractiveChartSelect(client).select_external_ids()

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=ChartExternalIdSelector(external_ids=tuple(selected_external_ids)),
                data=ChartIO(client),
                mapper=ChartMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def three_d(
        ctx: typer.Context,
        id: Annotated[
            list[int] | None,
            typer.Argument(
                help="The ID of the 3D Model to migrate. If not provided, an interactive selection will be "
                "performed to select the 3D Models to migrate."
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where migration logs will be stored.",
            ),
        ] = Path(f"migration_logs_{TODAY}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of "
                "what would be done is printed. This is useful for checking that all resources referenced by the 3D Models "
                "have been migrated to the new data modeling resources in CDF.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate 3D Models from Asset-Centric to data modeling in CDF.

        This command expects that the CogniteMigration data model is already deployed, and that the Mapping view
        is populated with the mapping from Asset-Centric resources to the new data modeling resources.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        selected_ids: list[int]
        if id:
            selected_ids = id
        else:
            selected_models = ThreeDInteractiveSelect(client, "migrate").select_three_d_models("classic")
            selected_ids = [model.id for model in selected_models]

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=ThreeDModelIdSelector(ids=tuple(selected_ids)),
                data=ThreeDMigrationIO(client),
                mapper=ThreeDMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def three_d_asset_mapping(
        ctx: typer.Context,
        model_id: Annotated[
            list[int] | None,
            typer.Argument(
                help="The IDs of the 3D model to migrate asset mappings for. If not provided, an interactive selection will be "
                "performed to select the."
            ),
        ] = None,
        object_3D_space: Annotated[
            str | None,
            typer.Option(
                "--object-3d-space",
                "-o",
                help="The instance space to ceate the 3D object nodes in.",
            ),
        ] = None,
        cad_node_space: Annotated[
            str | None,
            typer.Option(
                "--cad-node-space",
                "-c",
                help="The instance space to create the CAD node nodes in.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where migration logs will be stored.",
            ),
        ] = Path(f"migration_logs_{TODAY}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate 3D Model Asset Mappings from Asset-Centric to data modeling in CDF.

        This command expects that the selected 3D model has already been migrated to data modeling.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        selected_ids: list[int]
        if model_id is not None:
            selected_ids = model_id
        else:
            # Interactive selection
            selected_models = ThreeDInteractiveSelect(client, "migrate").select_three_d_models("dm")
            selected_ids = [model.id for model in selected_models]
            space_selector = DataModelingSelect(client, "migrate")
            object_3D_space = space_selector.select_instance_space(
                multiselect=False,
                message="In which instance space do you want to create the 3D Object nodes?",
                include_empty=True,
            )
            cad_node_space = space_selector.select_instance_space(
                multiselect=False,
                message="In which instance space do you want to create the CAD Node nodes?",
                include_empty=True,
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).unsafe_ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()

        if object_3D_space is None or cad_node_space is None:
            raise typer.BadParameter(
                "--object-3d-space and --cad-node-space are required when specifying IDs directly."
            )

        cmd = MigrationCommand(client=client)
        cmd.run(
            lambda: cmd.migrate(
                selected=ThreeDModelIdSelector(ids=tuple(selected_ids)),
                data=ThreeDAssetMappingMigrationIO(
                    client, object_3D_space=object_3D_space, cad_node_space=cad_node_space
                ),
                mapper=ThreeDAssetMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def infield_configs(
        ctx: typer.Context,
        external_id: Annotated[
            list[str] | None,
            typer.Argument(
                help="The external IDs of the APM Configurations to create Infield V2 configurations for. "
                "If not provided, an interactive selection will be performed to select the APM Configurations."
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the Infield V2 configuration definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates Infield V2 configurations from existing APM Configurations in CDF."""
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = MigrationCommand(client=client)
        if external_id is None:
            apm_configs = APMConfigInteractiveSelect(client, "migrate").select_apm_configs()
            output_dir = Path(
                questionary.path(
                    "Specify output directory for Infield V2 configuration definitions:", default=str(output_dir)
                ).unsafe_ask()
            )
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).unsafe_ask()
        else:
            apm_configs = None

        cmd.run(
            lambda: cmd.create(
                client,
                creator=InfieldV2ConfigCreator(client, external_id, apm_configs),
                output_dir=output_dir,
                dry_run=False,
                deploy=False,
                verbose=verbose,
            )
        )



import builtins
from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
)

from ._auth import BasicAuthenticationRequest, BasicAuthenticationResponse
from ._base import SourceRequestDefinition, SourceResponseDefinition


class EventHubSource(BaseModelObject):
    type: Literal["eventhub"] = "eventhub"
    host: str
    event_hub_name: str
    consumer_group: str | None = None


class EventHubSourceRequest(EventHubSource, SourceRequestDefinition):
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"authentication"})
    authentication: BasicAuthenticationRequest | None = None


class EventHubSourceResponse(
    SourceResponseDefinition,
    EventHubSource,
    ResponseResource[EventHubSourceRequest],
):
    authentication: BasicAuthenticationResponse | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[EventHubSourceRequest]:
        return EventHubSourceRequest



from typing import Any, Literal, TypeAlias

from pydantic import Field, field_validator
from pydantic_core.core_schema import ValidationInfo

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.utils import humanize_collection

from .base import BaseModelResource, ToolkitResource


class Disabled(BaseModelResource):
    enabled: Literal[False] = False


class ScheduleConfig(BaseModelResource):
    """Schedule configuration for the routine."""

    enabled: Literal[True] = True
    cron_expression: str = Field(description="Cron expression representing the schedule.")


class DataSamplingConfig(BaseModelResource):
    """Data sampling configuration for the routine."""

    enabled: Literal[True] = True
    validation_window: int | None = Field(
        default=None,
        description="Validation window of the data sampling. Represented in minutes. Used when either logical check or steady state detection is enabled.",
    )
    sampling_window: int = Field(description="Sampling window of the data sampling. Represented in minutes")
    granularity: int = Field(description="Granularity of the data sampling in minutes")


Aggregate = Literal["average", "interpolation", "stepInterpolation"]
LogicalCheckOperator = Literal["eq", "ne", "gt", "ge", "lt", "le"]


class LogicalCheckConfig(BaseModelResource):
    """Logical check configuration for validating time series data."""

    enabled: bool = Field(default=True, description="Whether this logical check is enabled.")
    timeseries_external_id: str = Field(description="External ID of the time series to check.")
    aggregate: Aggregate = Field(description="Aggregation method to apply.")
    operator: LogicalCheckOperator = Field(description="Comparison operator.")
    value: float = Field(description="Value to compare against.")


class SteadyStateDetectionConfig(BaseModelResource):
    enabled: bool = Field(default=True, description="Whether this steady state detection is enabled.")
    timeseries_external_id: str | None = Field(default=None, description="External ID of the time series to monitor.")
    aggregate: Aggregate = Field(description="Aggregation method to apply.")
    min_section_size: int = Field(description="Minimum section size for steady state detection.")
    var_threshold: float = Field(description="Variance threshold for steady state detection.")
    slope_threshold: float = Field(description="Slope threshold for steady state detection.")


ValueType: TypeAlias = Literal["STRING", "DOUBLE", "STRING_ARRAY", "DOUBLE_ARRAY"]


class Unit(BaseModelResource):
    """Unit definition for inputs and outputs."""

    name: str = Field(description="Name of the unit.", min_length=1, max_length=50)
    quantity: str = Field(description="Type of the unit.", min_length=1, max_length=50)


class RoutineInputConstantConfig(BaseModelResource):
    """Input configuration for the routine."""

    name: str = Field(description="Constant name")
    reference_id: str = Field(description="Reference ID for the input.")
    value: str | float | list[str] | list[float] = Field(description="Default value for the input.")
    value_type: ValueType = Field(default="STRING", description="Type of the value.")
    unit: Unit | None = Field(default=None, description="Unit of the input.")
    save_timeseries_external_id: str | None = Field(
        default=None, description="Time series external ID to use when saving the input sample in CDF."
    )


class RoutineInputTimeseriesConfig(BaseModelResource):
    """Input configuration for the routine."""

    name: str = Field(description="Timeseries name")
    reference_id: str = Field(description="Reference ID for the input.")
    source_external_id: str = Field(description="External id of the source time series to read from.")
    aggregate: Aggregate = Field(description="Aggregation method to apply.")
    unit: Unit | None = Field(default=None, description="Unit of the input.")
    save_timeseries_external_id: str | None = Field(
        default=None, description="Time series external ID to use when saving the input sample in CDF."
    )


class RoutineOutputConfig(BaseModelResource):
    """Output configuration for the routine."""

    name: str = Field(description="Name of the output.")
    reference_id: str = Field(description="Reference ID for the output.")
    unit: Unit | None = Field(default=None, description="Unit of the output.")
    value_type: ValueType = Field(default="STRING", description="Type of the value.")
    save_timeseries_external_id: str | None = Field(default=None, description="External ID for saving to time series.")


# ========== Main Configuration ==========
class SimulatorRoutineConfiguration(BaseModelResource):
    """Complete configuration for a simulator routine revision."""

    schedule: ScheduleConfig | Disabled = Field(description="Schedule configuration.", discriminator="enabled")
    data_sampling: DataSamplingConfig | Disabled = Field(description="Data sampling configuration.")
    logical_check: list[LogicalCheckConfig] = Field(
        description="List of logical checks.", default_factory=list, min_length=0, max_length=1
    )
    steady_state_detection: list[SteadyStateDetectionConfig] = Field(
        description="List of steady state detection configurations.", min_length=0, max_length=1, default_factory=list
    )
    inputs: list[RoutineInputConstantConfig] | None = Field(default=None, description="List of input configurations.")
    outputs: list[RoutineOutputConfig] | None = Field(default=None, description="List of output configurations.")


class ScriptStep(BaseModelResource):
    """A Get or Set step within a script stage."""

    order: int = Field(description="Order of the step within the stage.")
    description: str | None = Field(default=None, description="Description of the step.")
    step_type: Literal["Get", "Set", "Command"] = Field(description="Type of the step.")
    arguments: dict[str, Any] = Field(description="Arguments for the step.")

    @field_validator("arguments", mode="after")
    @classmethod
    def validate_arguments(cls, v: dict[str, Any], info: ValidationInfo) -> dict:
        step_type = info.data.get("step_type")
        if step_type not in ("Get", "Set"):
            return v  # No validation for other step types
        required_keys: set[str] = {"referenceId"}
        missing_keys = required_keys - set(v.keys())
        if missing_keys:
            raise ValueError(
                f"Missing required argument keys for step_type '{step_type}': {humanize_collection(missing_keys)}"
            )
        return v


class ScriptStage(BaseModelResource):
    """A stage in the script containing multiple steps."""

    order: int = Field(description="Order of the stage within the script.")
    description: str | None = Field(default=None, description="Description of the stage.")
    steps: list[ScriptStep] = Field(description="List of steps in this stage.")


class SimulatorRoutineRevisionYAML(ToolkitResource):
    """Simulator routine revision YAML resource class.

    Based on: https://api-docs.cognite.com/20230101/tag/Simulator-Routines/operation/create_simulator_routine_revision_simulators_routines_revisions_post
    """

    external_id: str = Field(description="External ID of the simulator routine revision.", min_length=1, max_length=255)
    routine_external_id: str = Field(description="External ID of the simulator routine.", min_length=1, max_length=255)
    configuration: SimulatorRoutineConfiguration = Field(description="Configuration of the simulator routine revision.")
    script: list[ScriptStage] | None = Field(default=None, description="Script stages for the routine revision.")

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



import difflib
from pathlib import Path
from typing import cast

import questionary
import typer
from questionary import Choice
from rich import print

from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.cruds import RESOURCE_CRUD_LIST, ResourceCRUD
from cognite_toolkit._cdf_tk.data_classes import ModuleDirectories
from cognite_toolkit._cdf_tk.resource_classes import ToolkitResource
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump


class ResourcesCommand(ToolkitCommand):
    def _get_or_prompt_module_path(self, module: str | None, organization_dir: Path, verbose: bool) -> Path:
        """
        Check if the module exists in the organization directory and return the module path.
        If module is not provided, ask the user to select or create a new module.
        """
        present_modules = ModuleDirectories.load(organization_dir, None)

        if module:
            for mod in present_modules:
                if mod.name.casefold() == module.casefold():
                    return mod.dir

            if questionary.confirm(f"{module} module not found. Do you want to create a new one?").unsafe_ask():
                return organization_dir / MODULES / module

            if verbose:
                print(f"[red]Aborting as {module} module not found...[/red]")
            else:
                print("[red]Aborting...[/red]")
            raise typer.Exit()

        choices = [Choice(title=mod.name, value=mod.dir) for mod in present_modules]
        choices.append(Choice(title="<Create new module>", value="NEW"))

        selected = questionary.select("Select a module:", choices=choices).unsafe_ask()

        if selected == "NEW":
            new_module_name = questionary.text("Enter name for new module:").unsafe_ask()
            if not new_module_name:
                print("[red]No module name provided. Aborting...[/red]")
                raise typer.Exit()
            return organization_dir / MODULES / new_module_name

        if not selected:
            print("[red]No module selected. Aborting...[/red]")
            raise typer.Exit()

        return cast(Path, selected)

    def _resolve_kinds(self, kinds: list[str] | None) -> list[type[ResourceCRUD]]:
        """
        Resolve kinds from list of strings or do an interactive selection.
        """
        all_cruds = {crud.kind.casefold(): crud for crud in RESOURCE_CRUD_LIST}

        if not kinds:
            sorted_cruds = sorted(RESOURCE_CRUD_LIST, key=lambda x: x.kind)
            choices = [Choice(title=crud.kind, value=crud) for crud in sorted_cruds]

            selected = questionary.select("Select resource type:", choices=choices).unsafe_ask()
            if not selected:
                print("[red]No resource type selected. Aborting...[/red]")
                raise typer.Exit()
            return [selected]

        resolved_cruds = []
        for kind in kinds:
            kind_lower = kind.casefold()
            if kind_lower in all_cruds:
                resolved_cruds.append(all_cruds[kind_lower])
            else:
                matches = difflib.get_close_matches(kind_lower, all_cruds.keys())
                if matches:
                    suggestion = all_cruds[matches[0]].kind
                    print(f"[red]Unknown resource type '{kind}'. Did you mean '{suggestion}'?[/red]")
                else:
                    print(
                        f"[red]Unknown resource type '{kind}'. "
                        f"Available types: {humanize_collection(sorted([c.kind for c in RESOURCE_CRUD_LIST]))}[/red]"
                    )
                raise typer.Exit()

        return resolved_cruds

    def _create_resource_yaml_skeleton(self, yaml_cls: type[ToolkitResource]) -> dict[str, str]:
        """
        Build YAML skeleton from a Pydantic model class using JSON schema for better type information.
        """
        yaml_skeleton: dict[str, str] = {}
        for field_name, field in yaml_cls.model_fields.items():
            name = field.alias or field_name
            description = field.description or name
            if field.is_required():
                yaml_skeleton[name] = f"(Required) {description}"
            else:
                yaml_skeleton[name] = description

        return yaml_skeleton

    def _get_resource_yaml_content(self, resource_crud: type[ResourceCRUD]) -> str:
        """
        Creates a new resource in the specified module using the resource_crud.yaml_cls.
        """
        yaml_header = (
            f"# API docs: {resource_crud.doc_url()}\n"
            f"# YAML reference: https://docs.cognite.com/cdf/deploy/cdf_toolkit/references/resource_library"
        )
        yaml_skeleton = self._create_resource_yaml_skeleton(resource_crud.yaml_cls)
        yaml_contents = yaml_safe_dump(yaml_skeleton)
        return yaml_header + "\n\n" + yaml_contents

    def _create_resource_yaml_file(
        self,
        resource_crud: type[ResourceCRUD],
        module_path: Path,
        prefix: str | None = None,
        verbose: bool = False,
    ) -> None:
        """
        Creates a new resource YAML file in the specified module using the resource_crud.yaml_cls.
        """
        resource_dir: Path = module_path / resource_crud.folder_name
        if resource_crud.sub_folder_name:
            resource_dir = resource_dir / resource_crud.sub_folder_name

        if not resource_dir.exists():
            resource_dir.mkdir(parents=True, exist_ok=True)

        final_prefix = prefix if prefix is not None else f"my_{resource_crud.kind}"
        file_name = f"{final_prefix}.{resource_crud.kind}.yaml"
        file_path: Path = resource_dir / file_name

        if (
            file_path.exists()
            and not questionary.confirm(f"{file_path.name} file already exists. Overwrite?").unsafe_ask()
        ):
            print("[red]Skipping...[/red]")
            return

        yaml_content = self._get_resource_yaml_content(resource_crud)
        file_path.write_text(yaml_content)
        if verbose:
            print(
                f"[green]{resource_crud.kind} Resource YAML file created successfully at {file_path.as_posix()}[/green]"
            )
        else:
            print(f"[green]Created {file_path.as_posix()}[/green]")

    def create(
        self,
        organization_dir: Path,
        module_name: str | None = None,
        kind: list[str] | None = None,
        prefix: str | None = None,
        verbose: bool = False,
    ) -> None:
        """
        create resource YAMLs.

        Args:
            organization_dir: The directory of the organization.
            module_name: The name of the module.
            kind: The kind(s) of resource to create.
            prefix: The prefix for the resource file.
            verbose: Whether to print verbose output.
        """
        module_path = self._get_or_prompt_module_path(module_name, organization_dir, verbose)
        resource_cruds = self._resolve_kinds(kind)
        for crud in resource_cruds:
            self._create_resource_yaml_file(crud, module_path, prefix, verbose)
